//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Hominid.java. (Model of an individual)                           *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  To do something.                                                 *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in Java to model proto-humans.                           *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.awt.Point;                                        // Java language.
import java.util.ArrayList;                                   //         ditto.
import java.util.Random;                                      //         ditto.

import root.hominini.ArtificialGenetics;               // program dependencies.
import root.hominini.Biology;                          //                ditto.
import root.hominini.HAL;                              //                ditto.
import root.hominini.HominidKit;                       //                ditto.
import root.hominini.Socioecology;                     //                ditto.
import root.ecology.Ecology;                           //                ditto.
import root.ecology.terrain.Terrain;                   //                ditto.
import root.events.LifeExperiences;                    //                ditto.
import root.ModelConstraints;                          //                ditto.
import root.rng.TrueRNG;                               //                ditto.

//
//*****************************************************************************
// Hominid

public class Hominid extends ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  private final float AGE_SKEW = 0.795F;         // likelihood age is 15 to 35.
  private final float AGE_PATHOLOGY = 0.625F;         // age related pathology.
  private final float FEM_RECEPTIVITY = 0.750F;  // F-only, prob mating occurs.

  // all time values are assumed to be in years unless otherwise noted.
  public  final int   MINIMUM_AGE = 6; // minimum age of an intial setup agent.
  private final int   TYPICAL_MIN = 15;  // typical min age intial setup agent.
  public  final int   MAXIMUM_AGE = 35;      // maximum age intial setup agent.
  private final boolean FEMALE_DISPERSAL = true;  // F-only, on maturity, walk.
  private final boolean MALE_DISPERSAL = false;   // M-only, on maturity, walk.

  private final int   MIN_DAYS_NO_WATER = 5; //minimum days any hominid can go.
  private final float WEIGHT_LOSS_DUE_TO_DISEASE = 0.010F;    // a scalar loss.
  private final int   LIFETIME_MAX_BIRTHS = 7; // F-only, 4-7 as max per Tutin.

  private final int   PREFERRED_GROUP_SIZE7  =  7;       // (Hill, 2003, p.68).
  private final int   PREFERRED_GROUP_SIZE21 = 21;         // works with gene6.
  private final int   PREFERRED_GROUP_SIZE35 = 35;         // works with gene6.

  // Variables.
  private long rng_Seed;                        // this is the GLOBAL RNG seed.
  private static int hLimit = 0;                    // horizontal grid surface.
  private static int vLimit = 0;                      // vertical grid surface.

  private String  causeOfDeath = "N/A";                // Age=A, C=catastrophy,
                             // lack of F=food, D=disease, IM=infant mortality,
             // N=needs reassignment, O=orphaned, P=predation, lack of W=water.
  private boolean extant = false;    // stored as boolean returns as yes or no.
  private Hominid mother = null;            // a pointer to the mother of self.
  private Hominid father = null;            // a pointer to the father of self.
  private Hominid child = null;    // a pointer to this hominid's child object.
  private long    agentID = 0L;       // system nanotime when chromosome built.
  private long    fatherID = 0L;      // system nanotime when chromosome built.
  private long    motherID = 0L;      // system nanotime when chromosome built.
  private Point   birthLoc = null;        // location where this agent started.
  private Point   currentLoc = null;           // location where this agent is.
  private Point   deathLoc = null;           // location where this agent died.
  private int     birthDate = 0;                // epoch when chromosome built.
  private int     deathDate = 0;          // epoch when chromosome deactivated.
  private String  preferredPrey = "";     // the prey currently most preferred.
  private String  rememberedPrey = "";                  // the prey remembered.
  private int     timesFoodSatisfies = 0;  // count of time food has satisfied.
  private int     maxDaysNoWater = 0;     // varies by chromosome/gene 1 value.
  private int     daysWithoutWater = 0;   // initially a hominid has had water.
  private boolean thirstDrive = false;     // level of thirst pleasure or pain.
  private boolean hungerDrive = false;     // level of hunger pleasure or pain.
  private boolean tempDrive = false;  // level of temperature pleasure or pain.
  private boolean socialDrive = false;     // level of social pleasure or pain.
  private boolean driveToDisperse = false; // level of reproductive motivation.
  private boolean panicDrive = false;      // level of fear sufficient to flee.
  private int     wentInWater = -9;    // reset event memory of going in water.
  private String  activeDrives = ""; // unary encoded, active, parallel drives.
  private Hominid benefactor = null; // an other giving stimulus reinforcement.
  private Hominid giftee = null;     // an other receiving antecident stimulus.

  private ArrayList <Hominid> siblings = null; //pntr to lvng siblings of self.
  private ArrayList <Hominid> children = null; //pntr to lvng children of self.
  private ArrayList <Hominid> inGroup = null;  // pntr to my ingroup of selves.

  // Objects.
  public ArtificialGenetics aGenes = null;        // pointer to the GA methods.
  public Biology bio = null;                   // pointer to biomimetic module.
  public HAL hal = null;        // pointer to heuristic, algorithmic, learning.
  public HominidKit kit = null;             // pointer to hominid supply cache.
  public LifeExperiences life = null;           // the daily lives of hominids.
  public Socioecology kin = null;            // pointer to relationship module.

  private static Hominid Adam = null;              // the father of all agents.
  private static Hominid Eve = null;
  private static Random RNG = null;          // pointer to standard RNG object.
  private static TrueRNG trueRNG = null;      // pointer to TrueRNG USB device.
  private static Random giftingRNG = null;   // pointer to Altruism RNG - ONLY.

  //***************************************************************************
  /** This is a class constructor.  It is a default (dummy) structure.*/

  public Hominid ()
  {

    // this is intentionally blank.  It is used by the agents started as seed.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This is a class constructor.  Phantom structure used seed agents.*/

  public Hominid (long newSeed, TrueRNG devicePntr)
  {

    this.rng_Seed = newSeed;            // make a local copy of the fixed seed.

    // if the TrueRNG USB device is available (ie devicePntr is not null) then,
    // the seed used by RNG will be truly random.  Otherwise, a fixed value of
    // newSeed will be used and all stochastic patterns will be deterministic.
    if (devicePntr == null)
      RNG = new Random(newSeed);            // an RNG object with a fixed seed.
    else
    {
      trueRNG = devicePntr;                // make a local copy of the pointer.
      RNG = new Random (devicePntr.getNextRandom()); // create with real rndom.
    }

    giftingRNG = new Random(newSeed);       // an RNG object with a fixed seed.

    setExtant(false);                                    // does not now exist.
    this.kin = new Socioecology (RNG.nextInt(), this);// pntr to kinship mthds.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is long method.

      This is a class constructor.  It makes adults roughly 15-35 years old.

      We make four assumptions here: 1) the greater likelihood is that a
      group member will be female.  This probability is set by CONSTANT; 2)
      there is greater likelihood that most hominids will be in the age
      the age range is 15 to 35, inclusive.  This probability is set by
      AGE_SKEW; 3) we assume the healthy weight of Hominid is set by a
      calculation based on an assumed, linear, absolute minimum sex
      differentiated "healthy weight" plus some uniform random additional
      amount;  4) we calculate and assign Hominid caloric and water
      requirements to each instance, mark them adult, and randomly assign
      pregnancy to females if they are older than an age CONSTANT.*/

  public Hominid (long newSeed,  TrueRNG devicePntr,
                  int epoch, Point location, int hLimit, int vLimit,
                  LifeExperiences virtualLife)
  {

    this.rng_Seed = newSeed;            // make a local copy of the fixed seed.

    // if the TrueRNG USB device is available (ie devicePntr is not null) then,
    // the seed used by RNG will be truly random.  Otherwise, a fixed value of
    // newSeed will be used and all stochastic patterns will be deterministic.
    if (devicePntr == null)
      RNG = new Random(newSeed);            // an RNG object with a fixed seed.
    else
      RNG = new Random(devicePntr.getNextRandom()); // create with real random.

    Adam = new Hominid (newSeed, devicePntr);      // the father of all agents.
    Eve = new Hominid (newSeed, devicePntr);       // the mother of all agents.

    this.kin = new Socioecology (RNG.nextInt(), this);// pntr to kinship mthds.
    this.bio = new Biology (RNG.nextInt(), this);     // pntr to biology mthds.
    this.aGenes = new ArtificialGenetics(RNG.nextInt()); // pntr to GA methods.
    this.life = virtualLife;  // pntr back into life experiences virtual world.

    // ------------------------------------------------------------------------
    setBirthLoc(location);                      // location provided by caller.
    setCurrentLoc(location);                    // location provided by caller.
    setDeathLoc(new Point(0, 0));                    // location of this agent.
    setAgentID(System.nanoTime());                           // this agents ID.
    setFatherID(0L);                   // the agent's geneological father's ID.
    setMotherID(0L);                   // the agent's geneological mother's ID.
    setBirthDate(epoch);                                   //agents birth date.
    bio.setChromosome(aGenes.makeChromosome());            // genetic metaphor.
    setExtant(true);                                             // now exists.
    setDeathDate(0);                                       //agents death date.
    // ------------------------------------------------------------------------
    // set this agent's age.
    float testAge = 0.0F;
    if (RNG.nextFloat() < AGE_SKEW)
    { // find a random age in the CONSTANT range TYPICAL_MIN to MAXIMUM_AGE.
      testAge = (RNG.nextFloat() *
                            ((float)(MAXIMUM_AGE - TYPICAL_MIN)))+ TYPICAL_MIN;
    }
    else
    { // find a random age in the CONSTANT range MINIMUM_AGE to TYPICAL_MIN.
      testAge = (RNG.nextFloat() *
                            ((float)(TYPICAL_MIN - MINIMUM_AGE)))+ MINIMUM_AGE;
    }
    bio.setAge ((int) testAge);                // cast final answer as integer.
    // ------------------------------------------------------------------------
    kin.setFather(Adam);             // set the pointer of this agent's father.
    kin.setMother(Eve);                 // set the pointer of this agent's mom.
    bio.setSex(aGenes.getSex(bio.getChromosome()));                     // set.
    // ------------------------------------------------------------------------
    // determine if this agent is a female then set fertility parameters.
    if (bio.getSex().equals("F"))
    {
      bio.setMObject(null);           // F-only, mated object, father of child.
      bio.setNewBirthNotice(false);  // initially all F are not about to birth.
      bio.setGestationDay(0);              // initially all F are not pregnant.
      bio.setPregnancyStatus(false);       // initially all F are not pregnant.
      bio.setNursingStatus(false);            // initially no F have offspring.
      bio.setLifetimeBirths(0);               // initially no F have offspring.
      if (bio.getAge() >= bio.getFemFertAge())
      {
        int newDate = RNG.nextInt(bio.FEM_NF_CYC_DAYS + bio.FEM_F_CYC_DAYS);
        bio.setFCycleDate(newDate);
      }
      else                                  // depending on age the cycle date.
        bio.setFCycleDate(0);
      bio.setFertile(false);                          // initially not fertile.
      bio.setHominidMaxBirthLimit(4 + RNG.nextInt(LIFETIME_MAX_BIRTHS - 3));
      // hominid maximum birth limit of birthings per Tutin, 1979, p.31.
    }
    // ------------------------------------------------------------------------
    // set this agent's age at weaning.
    bio.setWeaningAge();   // randomly set value in range (currently) 4 - 6 yr.
    // ------------------------------------------------------------------------
    // set this agent's preferred weight and fatal weight by linear relation.
    bio.setWeightPreference(bio.getIdealWeight (bio.getAge(),
                            bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
    bio.setCurrentWeight(bio.getWeightPreference()); // for now they are equal.
    // for starvation, disease pathology, and is adjusted by genetic influence.
    // survivable minimum living weight - uses gene 2.
    bio.setFatalWeight(bio.chromosome, ROBUST_METABOLISM_GENE);
    // ------------------------------------------------------------------------
    //discovery as B, J, or A then set it.
    bio.setHominidLifeStatus(bio.getLifeGroup());
    setCauseOfDeath("N/A");                    // not appropriate at this time.
    // ------------------------------------------------------------------------
    // set this agent's caloric and water needs.
    bio.setCaloricRequirement(bio.momentaryCaloricReq (bio.getAge(),
                              bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
    bio.setWaterRequirement (bio.momentaryWaterReq (bio.getAge(),
                             bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
    setDaysWithoutWater(0);       // initially, an agent should not be thirsty.
    setMaxDaysNoWater();        // varies according to chromosome/gene 1 value.
    setPreferredPrey("null");             // the prey currently most preferred.
    setRememberedPrey("null");                          // the prey remembered.
    setTimesFoodSatisfies(0); // count of times the current food has satisfied.
    // ------------------------------------------------------------------------
    // set this agent's pleasure and pain level.
    setHasPanic(false);         // initially agent has no panic (flight) drive.
    setIsThirsty(false);       // initially agent has no drive regarding water.
    setIsHungry(false);         // initially agent has no drive regarding food.
    setDriveToDisperse(false);   // start the level of reproductive motivation.
    setSocialDiscontent(false); // start agent comfortable in their social grp.
    setThermalDiscontent(false);     // start agent comfortable in the weather.
    this.activeDrives = initializeDrives(affect_word_width);     // set to 0's.
    // ------------------------------------------------------------------------
    // set this agent's Alpha status.
    this.kin.setIsAlpha(false);      // this hominid is currently not an Alpha.
    // ------------------------------------------------------------------------
    // start this agent's hominid supply of technology, food, and water.
    this.kit = new HominidKit (this.rng_Seed);             // hominid supplies.
    // ------------------------------------------------------------------------
    // start this agent's heuristic behavior suite.
    this.hal = new HAL (this.rng_Seed, this.currentLoc.x, this.currentLoc.y,
                        hLimit, vLimit);
    // ------------------------------------------------------------------------
    // set this agent's giftee and benefactor pointers to nulls initially.
    this.benefactor = null;
    this.giftee = null;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is long method.

      This is a class constructor.  It makes infants (0 year olds).

      We explicitly stipulate this is a baby.  A probability this hominid will
      be male is set by maleSkew.  This hominid has their weight set by
      CONSTANT.  Their caloric requirements are met by their motherand is
      0.0F.  More information is available in the "adult" version of this
      constructor.*/

  public Hominid (long newSeed, TrueRNG devicePntr,
                  Hominid mother, Hominid father, int epoch,
                  int hLimit, int vLimit, LifeExperiences virtualLife)
  {

    this.rng_Seed = newSeed;            // make a local copy of the fixed seed.

    // if the TrueRNG USB device is available (ie devicePntr is not null) then,
    // the seed used by RNG will be truly random.  Otherwise, a fixed value of
    // newSeed will be used and all stochastic patterns will be deterministic.
    if (devicePntr == null)
      RNG = new Random(newSeed);            // an RNG object with a fixed seed.
    else
      RNG = new Random(devicePntr.getNextRandom()); // create with real random.

    this.kin = new Socioecology (RNG.nextInt(), this);// pntr to kinship mthds.
    this.bio = new Biology (RNG.nextInt(), this);     // pntr to biology mthds.
    aGenes = new ArtificialGenetics(RNG.nextInt());    // link with GA methods.
    this.life = virtualLife;  // pntr back into life experiences virtual world.

    // ------------------------------------------------------------------------
    setBirthLoc(mother.currentLoc);              // location of agent's mother.
    setCurrentLoc(mother.currentLoc);                // location of this agent.
    Point deathLoc = new Point(0, 0);
    setDeathLoc(deathLoc);                           // location of this agent.
    setAgentID(System.nanoTime());                           // this agents ID.
    setFatherID(father.getAgentID());              // geneological father's ID.
    setMotherID(mother.getAgentID());              // geneological mother's ID.
    setBirthDate(epoch);                                   //agents birth date.
    bio.setChromosome(aGenes.reproduce(
                      father.bio.getChromosome(), mother.bio.getChromosome()));
    setExtant(true);                                             // now exists.
    setDeathDate(0);                                       //agents death date.
    // ------------------------------------------------------------------------
    bio.setAge (0);                      // explicitly assign baby age as zero.
    // ------------------------------------------------------------------------
    kin.setFather(father);           // set the pointer of this agent's father.
    kin.setMother(mother);              // set the pointer of this agent's mom.
    bio.setSex(aGenes.getSex(bio.getChromosome()));                     // set.
    // ------------------------------------------------------------------------
    // determine if this agent is a female then set fertility parameters.
    if (bio.getSex().equals("F"))
    {
      bio.setMObject(null);           // F-only, mated object, father of child.
      bio.setNewBirthNotice(false);  // initially all F are not about to birth.
      bio.setGestationDay(0);              // initially all F are not pregnant.
      bio.setPregnancyStatus(false);       // initially all F are not pregnant.
      bio.setNursingStatus(false);            // initially no F have offspring.
      bio.setLifetimeBirths(0);               // initially no F have offspring.
      bio.setFCycleDate(0);                              // initially set to 0.
      bio.setFertile(false);                          // initially not fertile.
      bio.setHominidMaxBirthLimit(4 + RNG.nextInt(LIFETIME_MAX_BIRTHS - 3));
      // hominid maximum birth limit of birthings per Tutin, 1979, p.31.
    }
    // ------------------------------------------------------------------------
    bio.setWeaningAge(); // a randomly set value in range (currently) 3 - 5 yr.
    // ------------------------------------------------------------------------
    bio.setWeightPreference(bio.BABY_WEIGHT);   // babies have constant weight.
    bio.setCurrentWeight(bio.BABY_WEIGHT);      // babies have constant weight.
    bio.setFatalWeight(bio.chromosome, ROBUST_METABOLISM_GENE);       // ditto.
    // ------------------------------------------------------------------------
    bio.setHominidLifeStatus ("B");              // set this hominid as a baby.
    setCauseOfDeath("N/A");                    // not appropriate at this time.
    // ------------------------------------------------------------------------
    bio.setCaloricRequirement(0.0F);  // the mother supplies all caloric needs.
    bio.setWaterRequirement (0.0F);     // the mother supplies all water needs.
    setDaysWithoutWater(0);                 // mother supplies all fluid needs.
    setMaxDaysNoWater();                    // mother supplies all fluid needs.
    setPreferredPrey("null");             // the prey currently most preferred.
    setRememberedPrey("null");                          // the prey remembered.
    setTimesFoodSatisfies(0); // count of times the current food has satisfied.
    // ------------------------------------------------------------------------
    // set this agent's pleasure and pain level.
    setHasPanic(false);         // initially agent has no panic (flight) drive.
    setIsThirsty(false);       // initially agent has no drive regarding water.
    setIsHungry(false);         // initially agent has no drive regarding food.
    setDriveToDisperse(false);   // start the level of reproductive motivation.
    setSocialDiscontent(false); // start agent comfortable in their social grp.
    setThermalDiscontent(false);     // start agent comfortable in the weather.
    this.activeDrives = initializeDrives(affect_word_width);     // set to 0's.
    // ------------------------------------------------------------------------
    // set this agent's Alpha status.
    this.kin.setIsAlpha(false);      // this hominid is currently not an Alpha.
    // ------------------------------------------------------------------------
    // start this agent's hominid supply of technology, food, and water.
    this.kit = new HominidKit (this.rng_Seed);             // hominid supplies.
    // ------------------------------------------------------------------------
    // start this agent's heuristic behavior suite.
    this.hal = new HAL (this.rng_Seed, this.currentLoc.x, this.currentLoc.y,
                        hLimit, vLimit);
    // ------------------------------------------------------------------------
    // set this agent's giftee and benefactor pointers to nulls initially.
    this.benefactor = null;
    this.giftee = null;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is a VERY, VERY long method.

      This method updates an Adult (and Juveniles by default) during an epoch.
      Factors under consideration include and in the order approached are:

      PANIC: movement that may result in survival.
      THIRST: movement to a location where water can be found.
      HUNGER: movement to a location where food can be found.
      GROUP: movement toward a location with an active cohort.
      SEXUAL DISPERSAL:  movement at sexual maturity, to find a mate.
      ENVIRONMENTAL STRESS:  movement to relieve environmental thermal stress.
      DISEASE:  random pathology may take away some of the hominid's weight.
      FEMALE REPRODUCTIVE CONTROLS:  reproduction related updates.
      MATING:  check if agent is female and fertile ...
      FEAR:  the mortal fear of drowning can create panic.
      FEAR:  the mortal fear of a predator can create panic.    (can be added)
      FEAR:  the mortal fear of the out-group can create panic. (can be added)
      SET REDUCIBLE DRIVES:  see if needs are being met.
      MORTALITY (DEHYDRATION, STARVATION, & AGE): "absolute" tests by trap.

      These are the critical functions of life that include a reduction of
      primitive drives.  These may result in relocation behaviors.*/

  public void adultLife (Ecology ecoSys, float ambientTemp, int epoch)
  {

    // ------------------------------------------------------------------------
    // begin by making copies of the cohort size, hominid age, and location.
    int cohortSize = ecoSys.groups.getCohortCount
                                   (getCurrentLoc().x, getCurrentLoc().y);
    int age = bio.getAge();                         // make a handy local copy.
    String sex = bio.getSex();                      // make a handy local copy.
    int h = getCurrentLoc().x;                     // the current local cell X.
    int v = getCurrentLoc().y;                     // the current local cell Y.
    hal.setCurrentXY(h, v);                          // you are currently here.

    // ------------------------------------------------------------------------
    // PANIC:  movement that may result in survival.

    if (this.activeDrives.substring(0, 1).equals("1"))
    {
      // There can be multiple panic causes and apparently panic now exists,
      // else this wouldn't be happening now.  However, if a move to reduce
      // the drive can be made then, clear the panic flag.  Else, if no move
      // can be made, leave the flag in place for LifeExperience.adultLife.
      setHasPanic (!hal.moveForFear(this, ecoSys,  fear_step));
      this.life.doRelocate (this, ecoSys, epoch);
    }

    // ------------------------------------------------------------------------
    // THIRST: movement to a location where water can be found.

    boolean attemptToFindWater = false;

    // if a residual drive from the last epoch exists, reduce it first.
    if (this.activeDrives.substring(1, 2).equals("1"))
    {
      hal.moveForThirst(this, ecoSys, water_step);    // look around for water.
      this.life.doRelocate (this, ecoSys, epoch);   // move tothe water source.
      attemptToFindWater = true;
    }
    h = getCurrentLoc().x;                         // the current local cell X.
    v = getCurrentLoc().y;                         // the current local cell Y.
    this.life.doForageWater (this, ecoSys, h, v);     // drink available water.
    float waterCollected = kit.getWaterOnHand();   // water that was collected.

    // DIAGNOSTIC
    // System.out.println("waterCollected 1: " + waterCollected);
    // DIAGNOSTIC

    float waterReq = bio.getWaterRequirement();   // rationrequirementsin gals.

    // DIAGNOSTIC
    // System.out.println("waterReq: " + waterReq);
    // DIAGNOSTIC

    if (waterCollected >= waterReq)          // was water collected sufficient.
    { // GREAT, there's enough for this hominid to drink.
      waterCollected -= waterReq;                      // take-away this drink.
      setDaysWithoutWater(0);                    // always reset after a drink.
      setIsThirsty(false);                        // agent thirst is satisfied.
    }
    else
    { // oops, there is too little to nothing to drink.
      if (!attemptToFindWater)
      {
        hal.moveForThirst(this, ecoSys, water_step);   // look aroundfor water.
        this.life.doRelocate (this, ecoSys, epoch);    // move to water source.

        h = getCurrentLoc().x;                     // the current local cell X.
        v = getCurrentLoc().y;                     // the current local cell Y.
        this.life.doForageWater (this, ecoSys, h, v);   // again try for water.
        waterCollected = kit.getWaterOnHand();   // whateverwaterwas collected.

        // DIAGNOSTIC
        // System.out.println("waterCollected 2: " + waterCollected);
        // DIAGNOSTIC

      }
      if (waterCollected >= waterReq)        // was water collected sufficient.
      { // there's enough for this hominid to drink.
        waterCollected -= waterReq;                    // take-away this drink.
        setDaysWithoutWater(0);                  // always reset after a drink.
        setIsThirsty(false);                      // agent thirst is satisfied.
      }
      else
      { // oops, there is little to nothing left.
        waterCollected = 0.0F;                             // zero is holdable.
        incrementDaysWithoutWater();                      // oops, this is bad.
        setIsThirsty(true);                   // agent thirst is not satisfied.
      }
    }
    kit.setWaterOnHand(waterCollected);   // put whateveris left back into kit.

    // DIAGNOSTIC
    // System.out.println("IsThirsty: " + getIsThirsty());
    // System.out.println("WaterOnHand: " + kit.getWaterOnHand());
    // DIAGNOSTIC

    // ------------------------------------------------------------------------
    // HUNGER: movement to a location where food can be found.

    /////////////////////////////////////////////////////////////////////
    // subtract one day's metabolic costs at current age and sex in kCal.
    float currentBodyWeightInCAL = bio.getCurrentWeight() *
                                   bio.FAT_AS_CALORIE;
    currentBodyWeightInCAL -= bio.getCaloricRequirement();
    bio.setCurrentWeight(currentBodyWeightInCAL / bio.FAT_AS_CALORIE);
    /////////////////////////////////////////////////////////////////////

    boolean attemptToFindFood = false;

    // if a residual drive from the last epoch exists, reduce it first.
    if (this.activeDrives.substring(2, 3).equals("1"))
    {
      hal.moveForHunger(this, ecoSys, food_step);      // look around for food.
      this.life.doRelocate (this, ecoSys, epoch);   // move to the food source.
      attemptToFindFood = true;
    }
    h = getCurrentLoc().x;                         // the current local cell X.
    v = getCurrentLoc().y;                         // the current local cell Y.
    this.life.doForageFood (this, ecoSys, h, v);         // collect avail food.

    float foodCollected = 0.0F;              // food that was collected in lbs.
    if (sex.equals("M"))
      foodCollected = kit.getFoodOnHand_M();                // a male foraging.
    else
      foodCollected = kit.getFoodOnHand_F();              // a female foraging.

    if (eatingHelper (foodCollected))            // consume the food collected.
    { // GREAT, there's enough for this hominid to eat.
      setIsHungry(false);                         // agent hunger is satisfied.
    }
    else
    { // oops, there is too little to nothing to eat.
      if (!attemptToFindFood)
      {
        hal.moveForHunger(this, ecoSys, food_step);    // look around for food.
        this.life.doRelocate (this, ecoSys, epoch);      // move to foodsource.

        h = getCurrentLoc().x;                     // the current local cell X.
        v = getCurrentLoc().y;                     // the current local cell Y.
        this.life.doForageFood (this, ecoSys, h, v);     // collect avail food.
        if (sex.equals("M"))
          foodCollected = kit.getFoodOnHand_M();            // a male foraging.
        else
          foodCollected = kit.getFoodOnHand_F();          // a female foraging.
      }

      if (eatingHelper (foodCollected))          // consume the food collected.
      { // there's enough for this hominid to eat.
        setIsHungry(false);                       // agent hunger is satisfied.
      }
      else
      {
        setIsHungry(true);                    // agent hunger is not satisfied.
      }
    }

    // ------------------------------------------------------------------------
    // GROUP: movement toward a location with an active cohort.

    if (this.activeDrives.substring(3, 4).equals("1"))
    {
      hal.moveForSocialGroup(this, ecoSys, group_step);
      this.life.doRelocate (this, ecoSys, epoch);
    }

    if (outsideSocialZone(cohortSize))
      setSocialDiscontent(true);      // drive to find a different sized group.
    else
      setSocialDiscontent(false);  // no drive to find a different sized group.

    // ------------------------------------------------------------------------
    // SEXUAL DISPERSAL:  movement at sexual maturity, to find a mate.

    if (this.activeDrives.substring(4, 5).equals("1"))
    {
      hal.moveForDispersal(this, ecoSys);
      this.life.doRelocate (this, ecoSys, epoch);
    }

    if (sexualDispersalRequest ())
      setDriveToDisperse (true);  // drive to locate a mate outside the cohort.
    else
      setDriveToDisperse (false);  // no drv to locate mate outside the cohort.

    // ------------------------------------------------------------------------
    // ENVIRONMENTAL STRESS:  movement to relieve environmental thermal stress.

    if (this.activeDrives.substring(7, 8).equals("1"))
    {
      hal.moveForTemp(this, ecoSys);
      this.life.doRelocate (this, ecoSys, epoch);
    }

    if (outsideThermalZone(ambientTemp))
      setThermalDiscontent(true);       // drive to find more comfortable zone.
    else
      setThermalDiscontent(false);   // no drive to find more comfortable zone.

    // ------------------------------------------------------------------------
    // DISEASE:  random pathology may take away some of the hominid's weight.

    if (RNG.nextFloat() < disease_threshold) // compare with disease threshold.
    {
      float tempWeight = bio.getCurrentWeight();     // remove a little weight.
      tempWeight = tempWeight - (tempWeight * WEIGHT_LOSS_DUE_TO_DISEASE);
      bio.setCurrentWeight(tempWeight);                    //  too bad, so sad.
    }

    // ------------------------------------------------------------------------
    // FEMALE REPRODUCTIVE CONTROLS:  reproduction related updates.

    if (sex.equals("F"))
      bio.updateReproductiveStatus (age, kin, MINIMUM_AGE, MAXIMUM_AGE,
                                    bio.chromosome, ROBUST_METABOLISM_GENE);

    // ------------------------------------------------------------------------
    // MATING:
    // check if agent is female and fertile, not pregnant, not nursing,
    // and is sexually mature, then by probability and genetic capacity
    // to effectively emit fertility scent (gene 6) do attempt to mate.

    h = getCurrentLoc().x;                         // the current local cell X.
    v = getCurrentLoc().y;                         // the current local cell Y.
    if ((sex.equals("F")) && (bio.getFertile()))
    {
      if ((! bio.getPregnancyStatus()) && (! bio.getNursingStatus()) &&
          (age >= bio.getFemFertAge()))
      {
        if ((RNG.nextFloat() - emitFertilityScent(cohortSize)) <
             FEM_RECEPTIVITY)
        {
          if (bio.getLifetimeBirths() <= bio.getHominidMaxBirthLimit())
          {
            this.life.mating (this, ecoSys.getCohort (h, v), epoch);
            // DIAGNOSTIC
            // System.out.println("Hominid.adultLife " + getAgentID () +
            //                    " presenting: " +
            //                    "......................................");
            // DIAGNOSTIC
          }
        }
      }
    }

    // ------------------------------------------------------------------------
    // FEAR:  the mortal fear of drowning can create panic.

    if (ecoSys.getBiomeName(getCurrentLoc().x,
                            getCurrentLoc().y).equals("water"))
      setHasPanic(true);                         // there is a cause for panic.

    // FEAR:  the mortal fear of a predator can create panic.
    // FEAR:  the mortal fear of out-group others can create panic.
    // add other FEAR code as needed.

    // ------------------------------------------------------------------------
    // SET REDUCIBLE DRIVES:  see if needs are being met.  If they are,
    // do nothing.  Else, react with prioritized, socio-spatial drives that
    // each reduce according to narrowly defined sensory-environmental
    // affect-effect interactions.

    setDrives (ecoSys);              // hominid must take some tough decisions.

    // ------------------------------------------------------------------------
    // MORTALITY (DEHYDRATION, STARVATION, & AGE):
    // test by trap.  If "underweight," "dehydrated," or very "old" either
    // sex can be declared dead if the bounds are exceeded.

    boolean dehydrated = (getDaysWithoutWater() > getMaxDaysNoWater());

    if (dehydrated)
    {
      //System.out.println("in Hominid.adultPhysiology -dehydrated- death");
      setCauseOfDeath ("W");      // death due to dehydration: lack of (W)ater.
      bio.setHominidLifeStatus("D");                          // set them dead.
      setExtant(false);                                      // reduce extancy.
    }

    if ( !bio.getHominidLifeStatus().equals("D") &&
          bio.getCurrentWeight() <= bio.getFatalWeight() )
    {
      //System.out.println("in Hominid.adultPhysiology -starving- death");
      setCauseOfDeath ("F");       // death due to lack of (F) food or wasting.
      bio.setHominidLifeStatus("D");                          // set them dead.
      setExtant(false);                                      // reduce extancy.
    }

    // this is a simple exponential trap.
    double deathKnell = 10.0D * Math.pow(bio.getAge(), -AGE_PATHOLOGY);
    if ( !bio.getHominidLifeStatus().equals("D") &&
         RNG.nextDouble() > deathKnell )
    {
      //System.out.println("in Hominid.adultPhysiology -old age- death");
      setCauseOfDeath("A");                       // death caused by (A) aging.
      bio.setHominidLifeStatus("D");                          // set them dead.
      setExtant(false);                                      // reduce extancy.
    }

    // ------------------------------------------------------------------------

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the capacity of the bearer to recognize kinship
      through paternal lineage.*/

  public void agnaticMemoryStructures ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is responsible for the expression of a gene that controls
      the tendency of the bearer to engage in socio-preferential movements.
      The gene is expressed similarly in males and females.  However, the age
      and sex of the bearer does differentiate bearer behavior especially
      during the year a female bearer becomes sexually mature.  Juvenile
      female bearers, during the year of the onset of their sexual maturity,
      will selectively choose random dispersal instead of preferential
      relocation to be with a benefactor.  At birth, the value of this gene is
      set by random Gaussian distribution over the integer interval 0 - 255,
      inclusive.

      Larger values of the gene lead the bearer to demonstrate the behavior
      of "selective co-location."  "Selective co-location" is defined as that
      condition that occurs when the bearer of the gene preferentially selects
      to relocate into a spatial cohort containing some other.  In particular,
      the other is one's benefactor*/

  public void alliance (Ecology ecoSys, int epoch)
  {

    final float SCALING = 0.00392F;             // results in 1.00 for 255 bin.
    final float ALLIANCE_ADVANTAGE = 0.50F; // arbitrary comparative advantage.

    // smaller gene values are less adaptive, larger genes are more adaptive.
    int geneInt = 0;                  // ranges from 0 - 255 in the 8-bit gene.
    geneInt = aGenes.getGeneValue (bio.chromosome, PAIR_BND_ALLIANCE_GENE);
    float geneticAdvantage = (float) geneInt * SCALING; // scaled as 0.0F-1.0F.

    if ((geneticAdvantage > ALLIANCE_ADVANTAGE) && (this.benefactor != null))
    {
      // observe and obey the drive for SEXUAL DISPERSAL at sexual maturity.
      if (!getDriveToDisperse())
      {
        // DIAGNOSTIC
        // System.out.println("Hominid.alliance moving to be near benefactor");
        // DIAGNOSTIC
        hal.moveForAlliance (this.benefactor);
        this.life.doRelocate (this, ecoSys, epoch);
      }
    }

    // ------------------------------------------------------------------------
    // SEXUAL DISPERSAL:  movement at sexual maturity, to find a mate.

    if (this.activeDrives.substring(4, 5).equals("1"))
    {
      hal.moveForDispersal(this, ecoSys);
      this.life.doRelocate (this, ecoSys, epoch);
    }

    if (sexualDispersalRequest ())
      setDriveToDisperse (true);  // drive to locate a mate outside the cohort.
    else
      setDriveToDisperse (false);  // no drv to locate mate outside the cohort.

    // ------------------------------------------------------------------------

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is responsible for the expression of a gene that controls
      the tendency of the bearer to engage in the "gifting" of excess foraging
      materials.  The gene expresses differentially in males and females.
      At birth, the value of this gene is set by random Gaussian distribution
      over the integer interval 0 - 255, inclusive.

      Differential sexual expression:
      Larger values of the gene lead the bearer to engage in the "gifting"
      of excess self-held food resources.  "Gifting" can only occur if the
      "gifting" agent has an excess of resource.
      a. In males:  "Gifting" may occur when the "gifting target" is female.
      b. In females:  "Gifting" may occur when the "gifting target" is the
      youngest, non-suckling offspring of the giver.*/

  public void altruism (Ecology ecoSys)
  {

    // AKA GIFTING...

    final float SCALING = 0.00392F;             // results in 1.00 for 255 bin.
    final float ALTRUISM_ADVANTAGE = 0.50F; // arbitrary comparative advantage.

    // smaller gene values are less adaptive, larger genes are more adaptive.
    int geneInt = 0;                  // ranges from 0 - 255 in the 8-bit gene.
    geneInt = aGenes.getGeneValue (bio.chromosome, SOCIAL_ALTRUISM_GENE);
    float geneticAdvantage = (float) geneInt * SCALING; // scaled as 0.0F-1.0F.
    if (geneticAdvantage > ALTRUISM_ADVANTAGE)
    {

      int locX = getCurrentLoc().x;       // this hominid's current X location.
      int locY = getCurrentLoc().y;       // this hominid's current Y location.
      ArrayList <Hominid> cohort = ecoSys.getCohort (locX, locY); // my cohort.

      // if you're not alone in this cell and if you're not hungry then
      // check if you're with a candidate giftee.  If your are do ...
      if ((cohort.size() > 1) && !getIsHungry())
      {

        // if you are an adult male do ...
        if (bio.getSex().equals("M") &&
            bio.getHominidLifeStatus().equals("A"))
        {
          this.giftee = getGiftee (cohort, "M");      // male gifting a female.
          // if you are with an eligible giftee give them food.
          if (this.giftee != null)
          {
            // if giftee is found, give her the food without regard to type.
            float foodInHandLBS = kit.getFoodOnHand_M(); // extra food, in lbs.
            life.benefitTransfer (this, this.giftee, foodInHandLBS);
            kit.setFoodOnHand_M(0.0F); // no food,remains, all was transferred.
          }
        }

        // if you are an adult female do ...
        if (bio.getSex().equals("F") &&
            bio.getHominidLifeStatus().equals("A"))
        {
          this.giftee = getGiftee (cohort, "F");  // female gifting a juvenile.
          if (this.giftee != null)
          {
            // is the juvenile giftee in the same cohort as its mother?
            if ((locX == this.giftee.getCurrentLoc().x) &&
                (locY == this.giftee.getCurrentLoc().y))
            {
              // if giftee is found, give them the food without regard to type.
              float foodInHandLBS = kit.getFoodOnHand_F();  // the food in lbs.
              life.benefitTransfer (this, this.giftee, foodInHandLBS);
              kit.setFoodOnHand_F(0.0F);           // all food was transferred.
            }
          }
        }
      }
    }

    /** Minimally, logical arguments in favor of a genetic basis for
        altruism began with - at least:
        Hamilton, W. D. (1963). The evolution of altruistic behavior. The
        American Naturalist, Vol. 97, No. 896 (Sep. - Oct., 1963),
        pp. 354-356.
        Hamilton, W. D. (1964). The genetical evolution of social behavior II.
        Journal of theoretical biology, Vol. 7. pp. 17-52.
        and was continued until - at least - by:
        Lehmann, L., & Keller, L. (2006). The evolution of cooperation and
        altruisma general framework and a classification of models. Journal
        of evolutionary biology, 19(5), 1365-1376.*/

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates a Baby during an epoch.*/

  public void babyLife (int epoch)
  {

    // if either as a consequence of random infant disease or that the mother
    // has died, record the cause of death and change the life status.
    if (RNG.nextFloat() < infanticide_mort)
    {
      setCauseOfDeath("IM");       // set death cause as (IM) infant mortality.
      bio.setHominidLifeStatus("D");                          // set them dead.
      setExtant(false);                                      // reduce extancy.
    }
    // ------------------------------------------------------------------------
    if (kin.getMother().bio.getHominidLifeStatus().equals("D"))
    {
      bio.setHominidLifeStatus("D");                          // set them dead.
      setCauseOfDeath("O");                    // death caused by (O) ophaning.
      setExtant(false);                                      // reduce extancy.
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method permits the execution of one or more short term behavioral
      control (re)actions.  These may be overt or covert.*/

  public void cognition (Ecology ecoSys, String stimuli)
  {

    if (stimuli.equals ( "Prey" ))
      hal.evalCogDissonance(this, ecoSys, stimuli);      // selects by stimuli.

    if (stimuli.equals ( "stub" ))
      hal.evalCogDissonance(this, ecoSys, stimuli);      // selects by stimuli.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is a helper to adultLife(HUNGER).  Because the process of
      eating involves taking from from the kit, converting mass to kCals,
      removing and or comparing required kCal to available food product then
      reconverting and or storing any residual mass as body fat, this is a
      complicated set of instructions.  foodCollected is whatever portion of
      food was foraged and foodKCalReq the momentary dietary requirement.*/

  public boolean eatingHelper (float foodInHandLBS)
  {

    // DIAGNOSTIC
    // System.out.println("foodInHandLBS: " + foodInHandLBS);
    // DIAGNOSTIC

    final float FAT_CAL = bio.FAT_AS_CALORIE; // nr of calories in a fat pound.
    boolean sated = false;        // assume meal will be inadequate to satiate.

    float idealWeightCAL = bio.getWeightPreference() * FAT_CAL;
    float currentBodyWeightInCAL = bio.getCurrentWeight() * FAT_CAL;
    float foodKCalReq = currentBodyWeightInCAL - idealWeightCAL;

    // DIAGNOSTIC
    // System.out.println("idealWeightCAL: " + idealWeightCAL);
    // System.out.println("currentBodyWeightInCAL: " + currentBodyWeightInCAL);
    // System.out.println("foodKCalReq: " + foodKCalReq);
    // DIAGNOSTIC

    if (foodKCalReq <= 0)
    { // you need to gain weight to be healthy, you have no fat reserves.
      foodKCalReq = Math.abs(foodKCalReq);
    }

    float mealKCalValue = 0.0F;               // this meal's actual KCal value.
    float foodInHandCAL = 0.0F;           // food in hand (collected), in kCal.
    foodInHandCAL = foodInHandLBS * FAT_CAL;              // cnvrt lbs to kCal.

    // DIAGNOSTIC
    // System.out.println("foodInHandCAL: " + foodInHandCAL);
    // DIAGNOSTIC

    if (foodInHandCAL >= foodKCalReq)     // remember, a kCal-based comparison.
    { // there's enough to feed this hominid.
      mealKCalValue = foodKCalReq;                 // yippee, go ahead and eat.
      foodInHandCAL -= foodKCalReq;    // subtract one meal from foods in hand.
      foodInHandLBS = foodInHandCAL / FAT_CAL;//don't overeat cnvrt rest to lb.
      sated = true;                                  // the meal was a success.
    }
    else
    { // oops, there was little to nothing to eat.
      mealKCalValue = foodInHandCAL;     // oops, the meal's meager kCal value.
      foodInHandLBS = 0.0F;         // in lbs., none remains, all was consumed.
      sated = false;                             // the meal was not a success.
    }
    if (bio.getSex().equals("M"))
      kit.setFoodOnHand_M(foodInHandLBS); // leftover food back in kit, in lbs.
    else
      kit.setFoodOnHand_F(foodInHandLBS); // leftover food back in kit, in lbs.

    // DIAGNOSTIC
    // System.out.println("mealKCalValue: " + mealKCalValue);
    // System.out.println("foodInHandLBS: " + foodInHandLBS);
    // DIAGNOSTIC

    // convert the net meal calorie value plus body weight back into lbs
    currentBodyWeightInCAL += mealKCalValue;
    bio.setCurrentWeight(currentBodyWeightInCAL / FAT_CAL);

    // DIAGNOSTIC
    // System.out.println("CurrentWeight: " + bio.getCurrentWeight());
    // System.out.println("sated: " + sated);
    // DIAGNOSTIC

    return sated;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the probability that a fertile, non-pregnant,
      non-nursing, sexually mature female will engage in mating behavior
      during estrus.  This gene expresses only in females.  At birth, the
      value of this gene is set by random Gaussian distribution over the
      integer interval 0 to 255, inclusive. Its purpose is to emulate the
      effect of a female sex pheromone whose likelihood of successful mating
      influence is directly proportionate to the size (numeric magnitude)
      of the cohort wherein the female is physically located.
      EMIT_FERTIL_SCENT relies on the ability of the female to recognize the
      number of individuals in her local small-group.  The resultant influence
      is graded against five numeric group size ranges which are 0, 1 - 7,
      8 - 21, 22 to 35, and more than 35.  A monotonically increasing
      likelihood of female presentation will occur in smaller numbered groups
      but not in larger.  Males cannot express this gene. This genetic trait
      is phenotypic.*/

  public float emitFertilityScent (int cohortSize)
  {

    final float GROUP_SIZE_7_ADVANTAGE = 500.0F;          // maximum advantage.
    final float GROUP_SIZE_21_ADVANTAGE = 1000.0F;        //  medium advantage.
    final float GROUP_SIZE_35_ADVANTAGE = 2000.0F;        // minimum advantage.

    int geneValue = 0;                // ranges from 0 - 255 in the 8-bit gene.
    geneValue = aGenes.getGeneValue(bio.chromosome, EMIT_FERTIL_SCENT_GENE);
    float geneticEquivalent = (float) geneValue;              // cast as float.

    // genetic advantage has three ranges with a maximum of 0.510 at stake.
    float geneticAdvantage = 0.0F; // genetic "pressure" to control group size.
    if (geneValue == 0)
      return 0.0F;                                // no advantage what so ever.
    else
    {
      if (cohortSize <= PREFERRED_GROUP_SIZE7)         // groups L.T. 7 agents.
        geneticAdvantage = GROUP_SIZE_7_ADVANTAGE;
      else if (cohortSize <= PREFERRED_GROUP_SIZE21)  // groups L.T. 21 agents.
        geneticAdvantage = GROUP_SIZE_21_ADVANTAGE;
      else                                            // groups G.T. 21 agents.
        geneticAdvantage = GROUP_SIZE_35_ADVANTAGE;

      // sawtooth output increases as gene values grow within each range but
      // overall, the output falls between ranges.
      // maximum output is 0.510 when the gene = 255 and the cohort = 7.
      return geneticEquivalent / geneticAdvantage;    // part of a probability.
    }

    // Adaptations to ovulation implications for sexual and social behavior,
    // Gangestad, S. W., Thornhill, R., and Garver-Apgar, C. E. (2005)
    // Current Directions in Psychological Science 14.6: pp. 312-316

    // Social Network Size in Humans. Hill,R.A. and Dunbar, R.I.M (2003)
    // Human Nature, Vol. 14, No. 1, pp. 53-72.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns this hominids activeDrives string.*/

  public String getActiveDrives ()
  {

    return this.activeDrives;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns this Agents ID.*/

  public long getAgentID ()
  {

    return this.agentID;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this agent's birth date.*/

  public int getBirthDate ()
  {

    return this.birthDate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this Hominids birth location.*/

  public Point getBirthLoc ()
  {

    return this.birthLoc;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the deceased's cause of death.*/

  public String getCauseOfDeath()
  {

    return this.causeOfDeath;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this Hominids currentlocation.*/

  public Point getCurrentLoc ()
  {

    return this.currentLoc;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's days without water.*/

  private int getDaysWithoutWater()
  {

    return this.daysWithoutWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this hominid's death date.*/

  public int getDeathDate ()
  {

    return this.deathDate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this hominid's death location.*/

  public Point getDeathLoc ()
  {

    return this.deathLoc;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the hominid's cause of death.*/
  public String getDeathReport()
  {

    if      (getCauseOfDeath().equals("IM"))
      return "Infant mortality";
    else if (getCauseOfDeath().equals("C"))
      return "Catastrophy";
    else if (getCauseOfDeath().equals("F"))
      return "Starvation";
    else if (getCauseOfDeath().equals("W"))
      return "Dehydration";
    else if (getCauseOfDeath().equals("A"))
      return "Age";
    else if (getCauseOfDeath().equals("O"))
      return "Orphaned";
    else if (getCauseOfDeath().equals("P"))
      return "Predation";
    else if (getCauseOfDeath().equals("S"))
      return "Drowning";
    else
     return "Disease";

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the agent's momentary drive to mating.*/

  public boolean getDriveToDisperse()
  {

    return this.driveToDisperse;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this agents status as extant or not.*/

  public boolean getExtant ()
  {

    return this.extant;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gets this agent's father ID.*/

  public long getFatherID ()
  {

    return this.fatherID;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method selects and returns a candidate giftee by caller sex.  To
      have arrived here you are not alone in your cohort, you are an adult,
      and your are not hungry.*/

  public Hominid getGiftee (ArrayList <Hominid> cohort, String callerSex)
  {

    Hominid giftee = null;          // initialize the giftee candidate as null.
    ArrayList <Integer> giftees = null;  // list of indices to eligible agents.

    // if you are an male do select an adult female giftee.
    if (callerSex.equals("M"))
    {
      int minAge = bio.getFemFertAge();        // minimum female fertility age.

      int females = life.countBySex(cohort, "F", minAge); // nr fems in cohort.
      if (females > 0)
      {
        // make a list of indices to sexually mature, candidate females.
        giftees = life.makeListByType (cohort, "F", minAge);
        if (giftees.size() > 0)           // if there is at least one eligible.
        {
          // randomly select one index to be the giftee.
          int i = giftingRNG.nextInt(giftees.size());
          Integer gifteeIndex = giftees.get(i);
          giftee = cohort.get(gifteeIndex.intValue());
        }
      }
    }
    giftees = null;                   // return memory to the operating system.

    // if you are an female with children do select your most recently weaned
    // offspring as giftee.  NOTE:  TEST for location of giftee is in caller!
    if (callerSex.equals("F") && kin.hasChildren())
    {
      Hominid child = null;
      int i = 0;
      boolean found = false;
      do
      {
        // starting with most recent born child ask ...
        child = kin.getChildren().get(i);
        if (child.bio.getHominidLifeStatus().equals("J"))
        {
          // DIAGNOSTIC
          // System.out.print("child-giftee identified with age: ");
          // System.out.println(child.bio.getAge());
          found = true;
        }
        else
          child = null;
      }
      while ((++i < kin.childrenCount()) && !found);
      giftee = child;
    }

    return giftee;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the agents momentary drive level regarding a
      fear sufficient to cause flight.*/

  public boolean getHasPanic ()
  {

    return this.panicDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the agents momentary drive regarding food.*/

  public boolean getIsHungry ()
  {

    return this.hungerDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the agents momentary drive regarding social groups.*/

  public boolean getIsSociallyContent ()
  {

    return this.socialDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the agents momentary drive regarding water.*/

  public boolean getIsThirsty ()
  {

    return this.thirstDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the hominid's life status.*/
  public String getLifeStatusReport()
  {

    String dummy = "ERROR MODE: in Hominid.getLifeStatusReport.";

    if (bio.getHominidLifeStatus().equals("A"))
      dummy =  "Adult";
    else if (bio.getHominidLifeStatus().equals("B"))
      dummy =  "Baby";
    else if (bio.getHominidLifeStatus().equals("J"))
      dummy =  "Juvenile";
    else if (bio.getHominidLifeStatus().equals("D"))
      dummy =  "Dead";
    else
      dummy = "ERROR MODE: in Homid.getLifeStatusReport.";

    return dummy;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's draught tolerance capacity in days.*/

  public int getMaxDaysNoWater ()
  {

   return this.maxDaysNoWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns this agent's mother's ID.*/

  public long getMotherID ()
  {

    return this.motherID;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the preferred food.*/

  public String getPreferredPrey()
  {

    return this.preferredPrey;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the remembered prey.*/

  public String getRememberedPrey()
  {

    return this.rememberedPrey;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the hominid's life status.*/
  public String getSexReport()
  {

    if (bio.getSex().equals("F"))
      return "Female";
    else
      return "Male";

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the agents momentary drive regarding social groups.*/

  public boolean getSocialDiscontent ()
  {

    return this.socialDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the agents momentary drive regarding temperature.*/

  public boolean getThermalDiscontent ()
  {

    return this.tempDrive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This methods gets the number of times a food has satisfied.*/

  public int getTimesFoodSatisfied()
  {

    return this.timesFoodSatisfies;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This methods gets the day number the individual went in the water.*/

  public int getWentInWater()
  {

    return this.wentInWater;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This increments the hominid's days without water.*/

  private void incrementDaysWithoutWater ()
  {

    int tempCount = this.getDaysWithoutWater();
    this.setDaysWithoutWater(++tempCount);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the activeDrives unary String to 0's.*/

  public String initializeDrives (int wordWidth )
  {

    int dummy = ((int) Math.pow( 2.0D, (double) wordWidth ));

    // convert the integer dummy a String mask of unary encoded
    // Binary bits.  The left-most "integer" bit is removed and the remainder
    // are converted into the returned String.
    return  Integer.toBinaryString(dummy).substring(1);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene expresses as an implicit violation of an unconscious social
      comfort zone.  It relies on the ability of the bearer to recognize the
      number of individuals in their local cohort. The bearer of this gene,
      whether male or female, experiences a monotonically increasing
      likelihood of discomfort when situated in larger social groups.*/

  public boolean outsideSocialZone (int cohortSize)
  {

    final float SCALING = 0.00392F;             // results in 1.00 for 255 bin.

    // smaller gene values are less adaptive, larger genes are more adaptive.
    int geneInt = 0;                  // ranges from 0 - 255 in the 8-bit gene.
    geneInt = aGenes.getGeneValue (bio.chromosome, EMIT_FERTIL_SCENT_GENE);
    float geneticAdvantage = (float) geneInt * SCALING; // scaled as 0.0F-1.0F.
    if ( geneInt == 0 ||
         geneticAdvantage <= 1.7F )
      geneticAdvantage = 1.0F;  // safety for zero division and clip the curve.
    else
      geneticAdvantage = 1.0F / geneticAdvantage; // small gene more advantage.

    float groupPreference = 0.0F;
    if (cohortSize == 0)                          // a cohort having no agents.
      groupPreference = 0.00001F;                              // no advantage.
    else if (cohortSize <= PREFERRED_GROUP_SIZE7)    // groups L.T.E. 7 agents.
      groupPreference = 0.99F;                               // best advantage.
    else if (cohortSize <= PREFERRED_GROUP_SIZE21)  // groups L.T.E. 21 agents.
      groupPreference = 0.29F;                            // better adavantage.
    else if (cohortSize <= PREFERRED_GROUP_SIZE35)  // groups L.T.E. 35 agents.
      groupPreference = 0.09F;                              // least advantage.
    else
      groupPreference = 0.00001F;                              // no advantage.

    if ( (geneticAdvantage * groupPreference) >= 0.50F)
      return true;                           // the status quo is unacceptable.
    else
      return false;                            // the status quo is acceptable.

    // "Social Network Size in Humans". Hill,R.A. and Dunbar, R.I.M (2003)
    // Human Nature, Vol. 14, No. 1, pp. 53-72.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method establishes in which range of temperatures an agent may be
      most comfortable.  It converts the numeric 0-255 code in gene site A
      into a 1-128 degree absolute temperature value.  It then establishes a
      nominal range of tempertures above and below which discomfort occurs.
      Any ambientTemp that falls within that range is returned true, meaning
      it is in the thermal comfort zone.  Those that fall outside the range
      are returned false.  The value of TEMPERATURE SETPOINT, is a genetic
      trait associated with temperature tolerance and is phenotypic.  Operates
      for either sex.*/

  public boolean outsideThermalZone (float ambientTemp)
  {

    final float STD_TEMP = 55.0F;   // a standard temperature ranged 55 - 79 F.
    final float UPPER_TOLERANCE = 10.0F;      // an upper temp tolerance limit.
    final float LOWER_TOLERANCE = 10.0F;       // a lower temp tolerance limit.

    int geneValue = 0;                // ranges from 0 - 255 in the 8-bit gene.
    geneValue = aGenes.getGeneValue(bio.chromosome, TEMPERTURE_SETPNT_GENE);
    float geneticEquivalent = (((float) geneValue) + 1.0F) / 23.0F;  // 0 - 11.
    geneticEquivalent = STD_TEMP +
                           geneticEquivalent * ( RNG.nextBoolean () ? -1: +1 );

    float upperLimit = geneticEquivalent + UPPER_TOLERANCE;         // nominal.
    float lowerLimit = geneticEquivalent - LOWER_TOLERANCE;           // ditto.

    if ((ambientTemp >= lowerLimit) && (ambientTemp <= upperLimit))
      return false;                         // no, this is in the thermal zone.
    else
      return true;                    // yes, this is outside the thermal zone.

    // "Physiological role of pleasure." Cabanac, M. (1971).
    // Science 173.4002: pp. 1103-1107.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates the hominid's life status.  It is assumed this method
      is called once each simulated day.  Further, it is assumed that a
      hominid's health life or death can be based solely on body weight.
      Because of the complexity involved in updating each type (B, J, A, D, N)
      of hominid, several helper functions are used.  For females, pregnancy
      and nursing cannot overlap.  Pregnancy is based on probability and
      after delivery results in nursing. This method uses helper funcions
      babyPhysiology and adultPhysiology.*/

  public void physiology (Ecology ecoSys, float ambientTemp, int epoch)
  {

    // -------------------------- HANDLE BIRTHDAYS ----------------------------
    // first, ensure today is not the agent's actual natal (birthing) day.
    if (epoch != getBirthDate())     // exclude the day of specific partuition.
    {

      if (((epoch - getBirthDate()) % YEAR) == 0) // is it an annular birthday.
      {
        // if the TrueRNG USB device is available (ie trueRNG is not null)
        // then refresh the RNG seed with a true random value.
        if (trueRNG != null)
          RNG.setSeed(trueRNG.getNextRandom());   // RNG seed with real random.

        int dummy = bio.getAge();                        // then increment age.
        bio.setAge(++dummy);

        // check if life status change is appropriate.
        if (bio.getAge() == bio.getWeaningAge())
        {
          bio.setHominidLifeStatus("J");  // congrats, you've just been weaned.
          // adopt mom's favorite forageables as your own.
          String momsPreferredPrey = kin.getMother().getPreferredPrey();
          String momsRememberedPrey = kin.getMother().getRememberedPrey();
          if (momsPreferredPrey.equals("none"))
            setPreferredPrey("fig");  // seed group mom started in "none" cell.
          else
            setPreferredPrey(momsPreferredPrey);  // mom had a real preference.
          setRememberedPrey(kin.getMother().getRememberedPrey()); //either way.

          // your body mass is now on the curve with all "J"s & "A"s.
          bio.setWeightPreference(bio.getIdealWeight (bio.getAge(),
                                  bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
          bio.setCurrentWeight(bio.getWeightPreference());

          // you've been with mom everyday since birth, correct the record.
          dummy = epoch - getBirthDate() - kin.contactDays(kin.getMother());
          for (int i = 0; i < dummy; i++)
            kin.updateGroup (kin.getMother()); // update mom in group memory.
        }
        // check if sexually mature and transition "J"s to "A"s.
        if (((bio.getSex().equals("F")) &&
             (bio.getAge() >= bio.getFemFertAge())) ||
            ((bio.getSex().equals("M")) &&
             (bio.getAge() >= bio.getMaleFertAge())))
          bio.setHominidLifeStatus("A");

        // from here on, J & A are the same but B, D, & N must be weeded out.
        String tmpLifeStatus = bio.getHominidLifeStatus();

        // once a year recalculate agent water, caloric, and fatality values.
        // NOTE: seed group instantiation, onset of F gestation, and B birth
        //       are handled elsewhere.  Some of these are redundant on
        //       instantiation and the "B" to "J" transition.
        if ((!tmpLifeStatus.equals("B")) && (!tmpLifeStatus.equals("D")))
        {
          bio.setWeightPreference(bio.getIdealWeight (bio.getAge(),
                                  bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
          bio.setWaterRequirement (bio.momentaryWaterReq (bio.getAge(),
                                   bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
          bio.setCaloricRequirement (bio.momentaryCaloricReq (bio.getAge(),
                                     bio.getSex(), MINIMUM_AGE, MAXIMUM_AGE));
          // survivable minimum living weight - uses gene 2.
          bio.setFatalWeight(bio.chromosome, ROBUST_METABOLISM_GENE);
        }

      }
    }
    // -------------------------- HANDLE BIRTHDAYS ----------------------------

    // ----------------- HANDLE PHYSIOLOGY BY LIFE STATUS ---------------------
    if (bio.getHominidLifeStatus().equals("B")) //check if hominid is (B) Baby.
      babyLife (epoch);                                     // update (B) homi.
    else if (bio.getHominidLifeStatus().equals("D"))//chck hominid (D) is dead.
    {
      System.out.println("ERROR: in Hominid.physiology condition (D)");
      System.exit(0); // error mode. Condition (D) is handled in LifeExperiens.
    }
    else if ((bio.getHominidLifeStatus().equals("A")) || //  homi is (A) adult.
             (bio.getHominidLifeStatus().equals("J"))) // homi is (J) juvenile.
       adultLife (ecoSys, ambientTemp, epoch);             // update (A) & (J).
    else
    {
      System.out.println("ERROR: in Hominid.adultPhysiology else failure");
      System.exit(0); // error mode.  Undefined agent state B, J, A, or D?
    }
    // ----------------- HANDLE PHYSIOLOGY BY LIFE STATUS ---------------------

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports status data on a specific hominid.  Used by the
      grimReaper() and others.*/

  public void report(int epoch)
  {

    System.out.println("####################################################");
    System.out.println("Hominid.report()\n");

    System.out.println("RNG seed:       " + this.rng_Seed);
    System.out.println("Birth date:     " + getBirthDate());
    System.out.println("Chromosome:     " + bio.getChromosome());
    System.out.println("Is extant:      " + (getExtant() ? "yes" : "no"));
    System.out.println("Agent ID:       " + getAgentID());
    System.out.println("Father ID:      " + getFatherID());
    System.out.println("Mother ID:      " + getMotherID());
    System.out.println("Birth location: " + this.birthLoc.x + "/" +
                                            this.birthLoc.y);
    System.out.println("Real location:  " + this.currentLoc.x + "/" +
                                            this.currentLoc.y);
    System.out.println("Death location: " + this.deathLoc.x + "/" +
                                            this.deathLoc.y);
    System.out.println("Death date:     " + getDeathDate());
    System.out.println("Age: (years)    " + bio.getAge());
    System.out.println("Age: (epoch)    " + (epoch - getBirthDate()));
    System.out.println("Sex:            " + getSexReport());
    System.out.println("Pregnant:       " + bio.getPregnancyStatus());
    Hominid h = bio.getMObject();
    System.out.println("Impregnator:    " + ( h!=null ? h.getAgentID():"N/A"));
    System.out.println("Nursing:        " + bio.getNursingStatus());
    System.out.println("P-Term date:    " + bio.getGestationDay());
    System.out.println("N-Term date:    " + bio.getNursingDay());
    System.out.println("Weight (lbs.):  " + bio.getCurrentWeight());
    System.out.println("Caloric Req:    " + bio.getCaloricRequirement());
    System.out.println("Water Req:      " + bio.getWaterRequirement());
    System.out.println("Age at weaning: " + bio.getWeaningAge());
    System.out.println("Individual is:  " + getLifeStatusReport());
    System.out.println("Was Alpha:      " + (kin.getIsAlpha() ? "yes" : "no"));
    System.out.println("Cause of death: " + getCauseOfDeath());
    reportChildren(this);               // reports hominids's list of children.
    reportSiblings(this);                // reports hominid's list of siblings.
    reportLocalGroup(this);   // reports hominid's list of local group members.

    System.out.println("\n");
    System.out.println("####################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports metabolic status data on a specific hominid.*/

  public void reportMetabolism ()
  {

    System.out.println("####################################################");
    System.out.println("Hominid.reportMetabolism()");
    System.out.println("RNG seed:          " + this.rng_Seed);
    System.out.println("Agent ID:          " + getAgentID());
    System.out.println("Age: (years)       " + bio.getAge());
    System.out.println("Age at weaning:    " + bio.getWeaningAge());
    System.out.println("Sex:               " + getSexReport());
    System.out.println("Current wt (lbs.): " + bio.getCurrentWeight());
    System.out.println("Fatal wt (lbs.):   " + bio.getFatalWeight());
    System.out.println("Caloric Req:       " + bio.getCaloricRequirement());
    System.out.println("Water Req:         " + bio.getWaterRequirement());
    System.out.println("####################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the hominid's children.*/

  public void reportChildren(Hominid p)
  {

    ArrayList <Hominid> kinder = p.kin.getChildren();

    System.out.print("My children:    ");
    if (kinder.size() == 0)
     System.out.println("N/A");
    else
    {

      for (int i = 0; i < kinder.size(); i++)
      {
        System.out.println(kinder.get(i).getAgentID());
        if (i < kinder.size() - 1)
          System.out.print("                ");
      }
    }

    siblings = null;                      // return memory to operating system.

  }
  // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports one hominids drives.*/

  public void reportDrives (Hominid hominid, String sMask)
  {

    System.out.println("####################################################");
    System.out.println("Hominid.reportDrives()\n");

    System.out.println("Hominid ID:            " + getAgentID());
    System.out.println(" -with trigger string: " + sMask);

    System.out.println (" hasPanic:         " +
                       (sMask.substring(0, 1).equals("1") ? true : false));
    System.out.println (" hasThirst:        " +
                       (sMask.substring(1, 2).equals("1") ? true : false));
    System.out.println (" hasHunger:        " +
                       (sMask.substring(2, 3).equals("1") ? true : false));
    System.out.println (" needsSocialGroup: " +
                       (sMask.substring(3, 4).equals("1") ? true : false));
    System.out.println (" needsDispersal:   " +
                       (sMask.substring(4, 5).equals("1") ? true : false));
    System.out.println (" undefined:        " +
                       (sMask.substring(5, 6).equals("1") ? true : false));
    System.out.println (" undefined:        " +
                       (sMask.substring(6, 7).equals("1") ? true : false));
    System.out.println (" hasDiscomfort:    " +
                       (sMask.substring(7, 8).equals("1") ? true : false));

    System.out.println("####################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports on the members of a hominid's local group.*/

  public void reportLocalGroup(Hominid p)
  {

    ArrayList <Hominid> group = p.kin.getLocalGroup();
    Hominid member = null;                 // a temporary pointer to a hominid.

    System.out.print("My local group: ");
    if (group.size() == 0)
     System.out.println("N/A");
    else
    {
      for (int i = 0; i < group.size(); i++)
      {
        member = group.get(i);
        System.out.println(member.getAgentID() + ", contact days: " +
                           p.kin.contactDays (member));
        if (i < group.size() - 1)
          System.out.print("                ");
      }
    }

    group = null;                         // return memory to operating system.
    member = null;                        // return memory to operating system.

  }
  // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the hominid's siblings.*/

  public void reportSiblings(Hominid p)
  {

    ArrayList <Hominid> siblings = p.kin.getSiblings();

    System.out.print("My siblings:    ");
    if (siblings.size() == 0)
     System.out.println("N/A");
    else
    {

      for (int i = 0; i < siblings.size(); i++)
      {
        System.out.println(siblings.get(i).getAgentID());
        if (i < siblings.size() - 1)
          System.out.print("                ");
      }
    }

    siblings = null;                      // return memory to operating system.

  }
  // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Agents ID.*/

  public void setAgentID (long newValue)
  {

    this.agentID = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Hominids benefactor property.*/

  public void setBenefactor (Hominid hominid)
  {
    this.benefactor = hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Hominids birth date.*/

  public void setBirthDate (int newTotal)
  {
    this.birthDate = newTotal;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Hominids birth location.*/

  public void setBirthLoc (Point newLocation)
  {

    this.birthLoc = newLocation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the deceased's cause of death.*/

  public void setCauseOfDeath (String deathCause)
  {

    this.causeOfDeath = deathCause;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Hominids currentlocation.*/

  public void setCurrentLoc (Point newLocation)
  {

    this.currentLoc = newLocation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's days without water.*/

  private void setDaysWithoutWater (int newDay)
  {

    this.daysWithoutWater = newDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this Hominids death date.*/

  public void setDeathDate (int newValue)
  {
    this.deathDate = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this hominid's death location.*/

  public void setDeathLoc (Point newLocation)
  {

    this.deathLoc = newLocation;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a Boolean signal announcing that the relocation of
      this Hominid agent has taken place.  The relocation corresponds to the
      result of testing a set of physiological, social, and cognitive drives
      according to a predetermined heuristic prioritization scheme.  If the
      respective trigger(s) are recognized (and there can be more than one
      trigger recognized) simultaneously, this method will return a Boolean
      true if the stimulus trigger caused the agent to respond with a physical
      relocation. A mask field is a representation of a binary switch field
      one bit longer than the true trigger field.  For example: a mask = 10000
      is a trigger field of four-bits.*/

  public void setDrives (Ecology ecoSys)
  {

    // ------------------------------------------------------------------------
    // drives sum in discrete prioritization order with exponential increments.

    int emptyMask = ((int) Math.pow( 2.0D, (double) affect_word_width ));
    Integer iMask = new Integer (emptyMask);

    // CURRENTLY (25-12-2014) 6 of 8-bits of trigger field are encoded.
    if (getHasPanic())                // is this individual currently in panic.
      iMask += 128;                                       // etc (8-bits long).

    if (getIsThirsty())                // is this individual currently thirsty.
      iMask += 64;                                        // etc (7-bits long).

    if (getIsHungry())                  // is this individual currently hungry.
      iMask += 32;                                        // etc (6-bits long).

    if (getIsSociallyContent())        // is this the right sized social group.
      iMask += 16;                                        // etc (5-bits long).

    if (getDriveToDisperse())           // is there active need to seek a mate.
      iMask += 8;                                         // etc (4-bits long).

    // values 4 and 2 are unused and "should" be available.

    if (getThermalDiscontent())            // is this environment thermally OK?
      iMask += 1;                                          // etc (1-bit long).

    // convert the Integer trigger mask to a String mask of unary encoded
    // Binary bits.  The left-most "integer" bit is removed and the remainder
    // are converted into the sMask String.
    String sMask = Integer.toBinaryString(iMask.intValue()).substring(1);

    // according to the numbeer GROUPED_DRIVES, leave only the left-most
    // GROUPED_DRIVES bits standing. Convert the rest (to the right) to 0's.
     this.activeDrives = hal.drivePrioritization (sMask, grouped_drives);

    // ------------------------------------------------------------------------
    // reportDrives (this, this.activeDrives);               // DIAGNOSTIC.
    // ------------------------------------------------------------------------

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agent's momentary drive to mating.*/

  public void setDriveToDisperse(boolean seekAMate)
  {

    this.driveToDisperse = seekAMate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this agents status as extant or not.*/

  public void setExtant (boolean newValue)
  {

    this.extant = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this agent's father ID.*/

  public void setFatherID (long newValue)
  {

    this.fatherID = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agents momentary drive level regarding a fear
      sufficient to cause flight.*/

  public void setHasPanic (boolean newValue)
  {

    this.panicDrive = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agents momentary drive regarding food.*/

  public void setIsHungry (boolean newPLevel)
  {

    this.hungerDrive = newPLevel;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agents momentary drive regarding water.*/

  public void setIsThirsty (boolean newPLevel)
  {

    this.thirstDrive = newPLevel;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method contributes to the effectiveness of agent draught tolerance
      If the value of gene site 1, DRAUGHT TOLERANCE, is expressed then the
      genetic trait is phenotypic.  Operates for either sex.*/

  public void setMaxDaysNoWater ()
  {

    final int ARBITRARY_ADVANTAGE = 1;          // arbitrary advantage of gene.

    int geneValue = 0;                // ranges from 0 - 255 in the 8-bit gene.
    geneValue = aGenes.getGeneValue(bio.chromosome, DRAUGHT_TOLERANCE_GENE);

    int result = 0;
    if (geneValue < 99)
      result = MIN_DAYS_NO_WATER + ARBITRARY_ADVANTAGE;      // gains one unit.
    else if (geneValue < 199)
      result = MIN_DAYS_NO_WATER + ARBITRARY_ADVANTAGE * 2;  // gains two unit.
    else
      result = MIN_DAYS_NO_WATER + ARBITRARY_ADVANTAGE * 3;      // gains unit.

   this.maxDaysNoWater = result;                           // set final result.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this agent's mother ID.*/

  public void setMotherID (long newValue)
  {

    this.motherID = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the preferred food.*/

  public void setPreferredPrey(String foodChoice)
  {

    this.preferredPrey = foodChoice;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method set the remembered prey.*/

  public void setRememberedPrey(String actualPrey)
  {

    this.rememberedPrey = actualPrey;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agents momentary drive regarding social groups.*/

  public void setSocialDiscontent (boolean newLevel)
  {

    this.socialDrive = newLevel;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the agents momentary drive regarding temperature.*/

  public void setThermalDiscontent (boolean newPLevel)
  {

    this.tempDrive = newPLevel;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This methods sets the number of times a food has satisfied.*/

  public void setTimesFoodSatisfies(int newCount)
  {

    this.timesFoodSatisfies = newCount;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This methods sets the day number the individual went in the water.*/

  public void setWentInWater(int theDay)
  {

    this.wentInWater = theDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the characteristics of host dispersal (leaving the
      natal community) versus philopatry (staying within the natal community)
      in the year the subject becomes sexually mature.  This gene can express
      in males, females, both, or neither depending on system experimental
      settings.  At birth, the value of this gene is set by random Gaussian
      distribution over the integer interval 0 - 255, inclusive. Only the
      integer value of PHILOPAT_DISPERSE, the subject sex, and the system
      experimental settings are involved.  Expression of this gene is
      phenotypic.*/

  public boolean sexualDispersalRequest ()
  {

    // Based on the absolute integer value of this gene, males and females
    // possessing integer values nearer to either 0 or 255 will be more
    // likely to leave their birth cohort to look for a mate rather than
    // staying.

    // @ +/-  7 ( 7, 247)
    // @ +/- 15 (15, 239)
    // @ +/- 31 (31, 224)
    // @ +/- 63 (63, 191)
    final int LOW_END = 7;
    final int HIGH_END = 247;
    boolean request = false;

    int geneValue = 0;                // ranges from 0 - 255 in the 8-bit gene.
    geneValue = aGenes.getGeneValue(bio.chromosome, PHILOPAT_DISPERSE_GENE);

    // test for extreme ends of gene ignoring the bigger middle part.
    if ((geneValue <= LOW_END) || (geneValue >= HIGH_END))
    {

      if ( FEMALE_DISPERSAL && (geneValue >= HIGH_END) )
      {
        // female, in the year of becoming sexually mature, and not pregnant.
        if ( bio.getSex().equals("F") && bio.getAge() == bio.getFemFertAge())
          if ( !bio.getPregnancyStatus() && !kin.hasChildren() )
            request = true;
          else
            request = false;
      }

      if ( MALE_DISPERSAL && (geneValue <= LOW_END) )
      {
        // if male and in the year of becoming sexually mature.
        if ( bio.getSex().equals("M") && bio.getAge() == bio.getMaleFertAge())
          if ( !kin.hasChildren() )
            request = true;
          else
            request = false;
      }
    }
    else
      request = false;

    return request;

    // The other "closest living relative": How bonobos (Pan paniscus)
    // challenge traditional assumptions about females, dominance, intra-
    // and intersexual dominance, and hominid evolution. Parish, A., and
    // De Waal, F. (2000). In LeCroy, D. & Moller, M. (Eds.), Annals of
    // the New York Academy of Sciences: Evolutionary perspectives on
    // human reproductive behavior, 907, pp. 97-113.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the likelihood the bearer will engage in
      potentially deadly territorial defense behaviors.*/

  public void territorialDefense ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the capacity of the bearer to recognize kinship
      through maternal lineage.*/

  public void uterineMemoryStructures ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
